"""Define the structure of a Pynecone app."""

import os
from collections import defaultdict
from typing import Any, Callable, Coroutine, Dict, List, Optional, Type, Union

import fastapi
from fastapi.middleware import cors

from pynecone import constants
from pynecone.base import Base
from pynecone.compiler.compiler import Compiler
from pynecone.components.component import Component, ComponentStyle
from pynecone.event import Event
from pynecone.middleware import HydrateMiddleware, LoggingMiddleware, Middleware
from pynecone.state import Delta, State

# Define custom types.
Page = Union[Component, Callable[[], Component]]
Reducer = Callable[[Event], Coroutine[Any, Any, Union[State, Delta]]]


class App(Base):
    """A Pynecone application."""

    # The directory to output the compiled components to.
    pages_root: str = constants.WEB_PAGES_DIR

    # A map from the path to the page.
    pages: Dict[str, Page] = {}

    # The backend api object.
    api: fastapi.FastAPI = None  # type: ignore

    # The state class to use for the app.
    state: Type[State] = State

    # The map from the client session token to the state.
    states: Dict[str, State] = {}

    # The styling to apply to each component.
    style: ComponentStyle = {}

    # Middleware to add to the app.
    middleware: List[Middleware] = [LoggingMiddleware(), HydrateMiddleware()]

    # The compiler to use.
    compiler: Compiler = Compiler()

    class Config:
        """Config to satisfy pydantic type checking."""

        arbitrary_types_allowed = True

    def __init__(self, *args, **kwargs):
        """Initialize the app.

        Args:
            *args: Args to initialize the app with.
            **kwargs: Kwargs to initialize the app with.
        """
        super().__init__(*args, **kwargs)
        self.states = defaultdict(self.state)
        self.api = fastapi.FastAPI()
        self.add_cors()
        self.add_default_endpoints()

    def add_default_endpoints(self):
        """Add the default endpoints."""
        # To test the server.
        self.get(constants.Endpoint.PING)(_ping)

        # To make state changes.
        self.post(constants.Endpoint.EVENT)(_event(app=self))

    def add_cors(self):
        """Add CORS middleware to the app."""
        self.api.add_middleware(
            cors.CORSMiddleware,
            allow_origins=["*"],
            allow_methods=["*"],
            allow_headers=["*"],
        )

    def get(self, path: Union[str, constants.Endpoint], *args, **kwargs) -> Callable:
        """Register a get request.

        Args:
            path: The endpoint path to link to the request.
            *args: Args to pass to the request.
            **kwargs: Kwargs to pass to the request.

        Returns:
            A decorator to handle the request.
        """
        return self.api.get(str(path), *args, **kwargs)

    def post(self, path: Union[str, constants.Endpoint], *args, **kwargs) -> Callable:
        """Register a post request.

        Args:
            path: The endpoint path to link to the request.
            *args: Args to pass to the request.
            **kwargs: Kwargs to pass to the request.

        Returns:
            A decorator to handle the request.
        """
        return self.api.post(str(path), *args, **kwargs)

    def preprocess(self, state: State, event: Event) -> Optional[State]:
        """Preprocess the event.

        Args:
            state: The state to preprocess.
            event: The event to preprocess.

        Returns:
            An optional state to return.
        """
        for middleware in self.middleware:
            out = middleware.preprocess(app=self, state=state, event=event)
            if out is not None:
                return out

    def postprocess(self, state: State, event: Event, delta: Delta) -> Optional[State]:
        """Postprocess the event.

        Args:
            state: The state to postprocess.
            event: The event to postprocess.
            delta: The delta to postprocess.

        Returns:
            An optional state to return.
        """
        for middleware in self.middleware:
            out = middleware.postprocess(
                app=self, state=state, event=event, delta=delta
            )
            if out is not None:
                return out

    def add_page(self, path: str, page: Page):
        """Add a page to the app.

        Args:
            path: The path to add the page to.
            page: The page to add.
        """
        # Format the path.
        path = path.strip(os.path.sep)
        if path == "":
            path = constants.INDEX
        self.pages[path] = page

    def compile(self):
        """Compile the app and output it to the pages folder."""
        for path, page in self.pages.items():
            # Generate the page if necessary.
            component = page if isinstance(page, Component) else page()
            assert isinstance(component, Component), "Expected a Component."

            # Add the style and state to the component.
            component.add_style(self.style).set_state(self.state)

            # Get the path for the output file.
            output_path = os.path.join(self.pages_root, path) + constants.JS_EXT
            code = self.compiler.compile(component=component, state=self.state)
            with open(output_path, "w") as f:
                f.write(code)


async def _ping() -> str:
    """Test API endpoint.

    Returns:
        The response.
    """
    return "pong"


def _event(app: App) -> Reducer:
    """Create an event reducer to modify the state.

    Args:
        app: The app to modify the state of.

    Returns:
        A handler that takes in an event and modifies the state.
    """

    async def process(event: Event) -> Union[State, Delta]:
        # Get the state for the session.
        state = app.states[event.token]

        # Preprocess the event.
        pre = app.preprocess(state, event)
        if pre is not None:
            return pre

        # Apply the event to the state.
        delta = state.process(event)

        # Postprocess the event.
        post = app.postprocess(state, event, delta)
        if post is not None:
            return post

        # Return the delta.
        return delta

    return process
