"""Basic component definitions."""

from __future__ import annotations

import inspect
import os
from abc import ABC
from collections import defaultdict
from typing import _GenericAlias  # type: ignore
from typing import Any, Dict, List, Set, Tuple, Type, Union

from pynecone import utils
from pynecone.base import Base
from pynecone.components.tag import Tag
from pynecone.event import EventChain, EventHandler, EventSpec
from pynecone.property import Property
from pynecone.state import State
from pynecone.style import Style

ImportDict = Dict[str, Set[str]]


class Component(Base, ABC):
    """The base class for all other pynecone components."""

    # The children nested within the component.
    children: List[Component] = []

    # An optional condition that must be True for the component to render.
    condition: Property[bool]

    # The style of the component.
    style: Style

    # The function to call when the component is clicked.
    on_click: EventChain

    # A unique key for the component.
    key: Any = None

    # The app state.
    state: Type[State] = State

    class Config:
        """Pydantic config."""

        arbitrary_types_allowed = True

    @classmethod
    def __init_subclass__(cls, **kwargs):
        """Set default properties.

        Args:
            **kwargs: The kwargs to pass to the superclass.
        """
        super().__init_subclass__(**kwargs)

        for field in cls.get_fields().values():
            # Only set for property types.
            if isinstance(field.type_, type) and issubclass(
                field.type_, (Property, EventSpec, EventChain)
            ):
                field.required = False
                field.default = Property.create(field.default)

    def __init__(self, *args, **kwargs):
        """Initialize the component.

        Args:
            *args: Args to initialize the component.
            **kwargs: Kwargs to initialize the component.
        """
        # Find the style attributes.
        if "style" not in kwargs:
            kwargs["style"] = Style()
        else:
            kwargs["style"] = kwargs["style"].copy()


        # Convert cond to condition.
        if "cond" in kwargs:
            kwargs["condition"] = kwargs.pop("cond")

        # Convert other fields to properties.
        fields = self.get_fields()
        for key, value in kwargs.items():
            # If the key is not in the fields, skip it.
            if key not in fields:
                continue

            # If the field does not have a type, skip it.
            if not inspect.isclass(fields[key].type_):
                continue

            field_type = fields[key].type_

            # Convert event handlers into event specs.
            if issubclass(field_type, EventChain) and not isinstance(value, EventChain):
                if not isinstance(value, utils.get_args(Union[List, Set, Tuple])):
                    value = [value]
                value = [
                    event() if isinstance(event, EventHandler) else event
                    for event in value
                ]
                kwargs[key] = EventChain(events=value)

            # Check whether the key is a component prop.
            if issubclass(field_type, Property):
                # Convert any constants into properties.
                kwargs[key] = Property.create(value)

                # Make sure the property types match.
                expected_type = fields[key].outer_type_.__args__[0]
                if isinstance(expected_type, _GenericAlias):
                    expected_type = expected_type.__origin__
                assert issubclass(
                    kwargs[key].type_, expected_type
                ), f"Invalid property passed for {key}, expected {expected_type}, got {kwargs[key].type_}."

        kwargs["style"].update(
            {
                utils.to_camel_case(attr): value
                for attr, value in kwargs.items()
                if attr not in self.get_fields().keys()
            }
        )
        super().__init__(*args, **kwargs)

    def __repr__(self) -> str:
        """Represent the component in React.

        Returns:
            The code to render the component.
        """
        return self.render()

    def __str__(self) -> str:
        """Represent the component in React.

        Returns:
            The code to render the component.
        """
        return self.render()

    def _render(self) -> Tag:
        """Define how to render the component in React."""
        raise NotImplementedError

    @classmethod
    def create(cls, *children, **attrs) -> Component:
        """Create the component.

        Args:
            *children: The children of the component.
            **attrs: The attributes of the component.

        Returns:
            The component.
        """
        from pynecone.components.typography.text import Text

        children = [
            Text.create(text=child, as_="span")
            if not isinstance(child, Component)
            else child
            for child in children
        ]
        return cls(children=children, **attrs)

    def add_style(self, style: ComponentStyle) -> Component:
        """Add additional style to the component and its children.

        Args:
            style: A dict from component to styling.

        Returns:
            The component with the additional style.
        """
        if type(self) in style:
            self.style.update(style[type(self)])
        for child in self.children:
            child.add_style(style)
        return self

    def set_state(self, state: Type[State]) -> Component:
        """Set the state of the component.

        Args:
            state: The state to set.

        Returns:
            The component with the state set.
        """
        self.state = state
        for child in self.children:
            child.set_state(state)
        return self

    def render(self) -> str:
        """Render the component.

        Returns:
            The code to render the component.
        """
        tag = self._render()
        return str(
            tag.add_attrs(on_click=self.on_click, key=self.key, sx=self.style).set(
                condition=self.condition,
                contents=os.linesep.join(
                    [str(tag.contents)] + [child.render() for child in self.children]
                ),
            )
        )

    def _get_imports(self) -> ImportDict:
        return {}

    def get_imports(self) -> ImportDict:
        """Get all the libraries and fields that are used by the component.

        Returns:
            The import dict with the required imports.
        """
        return merge_imports(
            self._get_imports(), *[child.get_imports() for child in self.children]
        )


def merge_imports(*imports) -> ImportDict:
    """Merge two import dicts together.

    Args:
        *imports: The list of import dicts to merge.

    Returns:
        The merged import dicts.
    """
    all_imports = defaultdict(set)
    for import_dict in imports:
        for lib, fields in import_dict.items():
            for field in fields:
                all_imports[lib].add(field)
    return all_imports


# Map from component to styling.
ComponentStyle = Dict[Type[Component], Style]
