"""An editable component."""

from typing import Any, Tuple

from pynecone.components.chakra import ChakraComponent
from pynecone.components.mixins import ControlledMixin
from pynecone.components.tag import Tag
from pynecone.event import EventChain
from pynecone.property import Property


class Editable(ControlledMixin, ChakraComponent):
    """The wrapper component that provides context value."""

    tag = "Editable"

    # If true, the Editable will be disabled.
    is_disabled: Property[bool]

    # If true, the read only view, has a tabIndex set to 0 so it can receive focus via the keyboard or click.
    is_preview_focusable: Property[bool]

    # The placeholder text when the value is empty.
    placeholder: Property[str]

    # If true, the input's text will be highlighted on focus.
    select_all_on_focus: Property[bool]

    # If true, the Editable will start with edit mode by default.
    start_with_edit_view: Property[bool]

    # If true, it'll update the value onBlur and turn off the edit mode.
    submit_on_blur: Property[bool]

    # The value of the Editable in both edit & preview mode
    value: Property[str]

    # The type of Editable.
    type_: Property[str] = "text"  # type: ignore

    # The initial value of the Editable in both edit and preview mode.
    default_value: Property[str]

    # Callback invoked when user cancels input with the Esc key. It provides the last confirmed value as argument.
    on_cancel: EventChain

    # Callback invoked once the user enters edit mode.
    on_edit: EventChain

    # Callback invoked when user confirms value with enter key or by blurring input.
    on_submit: EventChain

    def _get_local_args(self) -> Tuple[str, ...]:
        return ("val",)

    def _get_handler_args(self) -> Tuple[Tuple[str, Any], ...]:
        return ((self.value.name, "val"),)

    def _render(self) -> Tag:
        return (
            super()
            ._render()
            .add_attrs(
                type=self.type_,
            )
        )


class EditableInput(ChakraComponent):
    """The edit view of the component. It shows when you click or focus on the text."""

    tag = "EditableInput"


class EditableTextarea(ChakraComponent):
    """Use the textarea element to handle multi line text input in an editable context."""

    tag = "EditableTextarea"


class EditablePreview(ChakraComponent):
    """The read-only view of the component."""

    tag = "EditablePreview"
