"""An input component."""

from pynecone import utils
from pynecone.components.chakra import ChakraComponent
from pynecone.components.mixins import ControlledMixin
from pynecone.components.tag import Tag
from pynecone.property import Property


class Input(ControlledMixin, ChakraComponent):
    """The Input component is a component that is used to get user input in a text field."""

    tag = "Input"

    # State property to bind the the input.
    value: Property[str]

    # The placeholder text.
    placeholder: Property[str]

    # The type of input.
    type_: Property[str] = "text"  # type: ignore

    # The id of the input.
    id: Property[str]

    # The border color when the input is invalid.
    error_border_color: Property[str]

    # The border color when the input is focused.
    focus_border_color: Property[str]

    # If true, the form control will be disabled. This has 2 side effects - The FormLabel will have `data-disabled` attribute - The form element (e.g, Input) will be disabled
    is_disabled: Property[bool]

    # If true, the form control will be invalid. This has 2 side effects - The FormLabel and FormErrorIcon will have `data-invalid` set to true - The form element (e.g, Input) will have `aria-invalid` set to true
    is_invalid: Property[bool]

    # If true, the form control will be readonly.
    is_read_only: Property[bool]

    # If true, the form control will be required. This has 2 side effects - The FormLabel will show a required indicator - The form element (e.g, Input) will have `aria-required` set to true
    is_required: Property[bool]

    # "outline" | "filled" | "flushed" | "unstyled"
    variant: Property[str]

    @property
    def default_placeholder(self) -> str:
        """Get the default placeholder text.

        Returns:
            The default placeholder text.
        """
        if self.value is None:
            return ""
        return utils.to_title(self.value.name)

    def _render(self) -> Tag:
        return (
            super()
            ._render()
            .add_attrs(
                placeholder=self.placeholder
                if self.placeholder is not None
                else self.default_placeholder,
                type=self.type_,
            )
        )


class InputGroup(ChakraComponent):
    """The InputGroup component is a component that is used to group a set of inputs."""

    tag = "InputGroup"


class InputLeftAddon(ChakraComponent):
    """The InputLeftAddon component is a component that is used to add an addon to the left of an input."""

    tag = "InputLeftAddon"


class InputRightAddon(ChakraComponent):
    """The InputRightAddon component is a component that is used to add an addon to the right of an input."""

    tag = "InputRightAddon"
