"""A React tag."""

from __future__ import annotations

import json
import os
import re
from typing import Any, Dict, Optional, Union

import pydantic

from pynecone import utils
from pynecone.base import Base
from pynecone.event import EventChain, EventSpec
from pynecone.property import Property
from pynecone.style import Style


class Tag(Base):
    """A React tag."""

    # The name of the tag.
    name: str = ""

    # The attributes of the tag.
    attrs: Dict[str, Any] = {}

    # The inner contents of the tag.
    contents: str = ""

    # The condition that must be True for the tag to render.
    condition: Optional[Property[bool]] = None

    class Config:
        """Pydantic config."""

        arbitrary_types_allowed = True

    def __init__(self, *args, **kwargs):
        """Initialize the tag.

        Args:
            *args: Args to initialize the tag.
            **kwargs: Kwargs to initialize the tag.
        """
        # Convert any attrs to properties.
        if "attrs" in kwargs:
            kwargs["attrs"] = {
                name: Property.create(value) for name, value in kwargs["attrs"].items()
            }
        super().__init__(*args, **kwargs)

    @pydantic.validator("condition")
    def validate_condition(cls, condition: Property) -> Property:
        """Validate that the condition is a boolean.

        Args:
            condition: The condition to validate.

        Returns:
            The validated condition.
        """
        assert issubclass(condition.type_, bool), "The property must be a boolean."
        return condition

    @staticmethod
    def format_attr_value(
        value: Union[Property, EventSpec, EventChain, Dict[str, Property], str],
    ) -> Union[int, float, str]:
        """Format an attribute value.

        Args:
            value: The value of the attribute

        Returns:
            The formatted value to display within the tag.
        """

        def format_fn(value):
            args = ",".join([":".join((name, val)) for name, val in value.args])
            return f"E(\"{utils.to_snake_case(value.handler.fn.__qualname__)}\", {utils.wrap(args, '{')})"

        # Handle property attributes.
        if isinstance(value, Property):
            if not value.is_local:
                return str(value)
            if issubclass(value.type_, str):
                return json.dumps(value.full_name)
            value = value.full_name

        # Handle events.
        elif isinstance(value, EventSpec):
            local_args = ",".join(value.local_args)
            fns = format_fn(value)
            value = f"({local_args}) => Event([{fns}])"

        elif isinstance(value, EventChain):
            fns = ",".join([format_fn(event) for event in value.events])
            value = value.events[0]
            local_args = ",".join(value.local_args)
            value = f"({local_args}) => Event([{fns}])"

        # Handle other types.
        elif isinstance(value, str):
            if utils.is_wrapped(value, "{"):
                return value
            return json.dumps(value)

        # For dictionaries, convert any properties to strings.
        else:
            if isinstance(value, dict):
                value = json.dumps(
                    {
                        key: str(val) if isinstance(val, Property) else val
                        for key, val in value.items()
                    }
                )
            else:
                value = json.dumps(value)

            value = re.sub('"{', "", value)
            value = re.sub('}"', "", value)

        # Wrap the variable in braces.
        return utils.wrap(value, "{", check_first=False)

    def format_attrs(self) -> str:
        """Format a dictionary of attributes.

        Returns:
            The formatted attributes.
        """
        # If there are no attributes, return an empty string.
        if len(self.attrs) == 0:
            return ""

        # Get the string representation of all the attributes joined.
        # We need a space at the beginning for formatting.
        return os.linesep.join(
            f"{name}={self.format_attr_value(value)}"
            for name, value in self.attrs.items()
            if value is not None
        )

    def __str__(self) -> str:
        """Render the tag as a React string.

        Returns:
            The React code to render the tag.
        """
        # Get the tag attributes.
        attrs_str = self.format_attrs()
        if len(attrs_str) > 0:
            attrs_str = " " + attrs_str

        if len(self.contents) == 0:
            # If there is no inner content, we don't need a closing tag.
            tag_str = utils.wrap(f"{self.name}{attrs_str}/", "<")
        else:
            # Otherwise wrap it in opening and closing tags.
            open = utils.wrap(f"{self.name}{attrs_str}", "<")
            close = utils.wrap(f"/{self.name}", "<")
            tag_str = utils.wrap(utils.indent(self.contents), open, close)

        # If there is a condition, we must conditionally render the tag.
        if self.condition is not None:
            tag_str = utils.format_conditional(self.condition.full_name, tag_str)

        return tag_str

    def add_attrs(self, **kwargs: Optional[Any]) -> Tag:
        """Add attributes to the tag.

        Args:
            **kwargs: The attributes to add.

        Returns:
            The tag with the attributes added.
        """
        self.attrs.update(
            {
                utils.to_camel_case(name): attr
                if isinstance(attr, utils.get_args(Union[EventSpec, EventChain, Style]))
                else Property.create(attr)
                for name, attr in kwargs.items()
                if self.is_valid_attr(attr)
            }
        )
        return self

    @staticmethod
    def is_valid_attr(attr: Optional[Property]) -> bool:
        """Check if the attr is valid.

        Args:
            attr: The value to check.

        Returns:
            Whether the value is valid.
        """
        return attr is not None and not (isinstance(attr, dict) and len(attr) == 0)
