"""Define the event specification."""
from __future__ import annotations

import inspect
from datetime import datetime
from typing import Any, Callable, Dict, List, Tuple

import pydantic

from pynecone.base import Base
from pynecone.property import Property


class Event(Base):
    """An event that describes any state change in the app."""

    # The token to specify the client that the event is for.
    token: str

    # The event name.
    name: str

    # The event payload.
    payload: Dict[str, Any] = {}

    # The date the event was created.
    date: datetime = pydantic.Field(default_factory=datetime.now)


class EventHandler(Base):
    """An event handler responds to an event to update the state."""

    # The function to call in response to the event.
    fn: Callable

    class Config:
        """The Pydantic config."""

        # Needed to allow serialization of Callable.
        frozen = True

    def __call__(self, *args: Property) -> EventSpec:
        """Pass arguments to the handler to get an event spec.

        This method configures event handlers that take in arguments.

        Args:
            *args: The arguments to pass to the handler.

        Returns:
            The event spec, containing both the function and args.
        """
        # Make sure the args passed in match the function signature.
        fn_args = inspect.getfullargspec(self.fn).args[1:]

        # TODO(nikhil): The assert should be valid even when len == 0.
        if len(args) > 0:
            assert len(args) == len(
                fn_args
            ), f"Expected {len(fn_args)} arguments, got {len(args)}"
            payload = tuple(zip(fn_args, [arg.full_name for arg in args]))
        else:
            payload = tuple()

        # Construct the event payload.
        return EventSpec(handler=self, args=payload)


class EventSpec(Base):
    """An event specification.

    Whereas an Event object is passed during runtime, a spec is used
    during compile time to outline the structure of an event.
    """

    # The event handler.
    handler: EventHandler

    # The local arguments on the frontend.
    local_args: Tuple[str, ...] = ()

    # The arguments to pass to the function.
    args: Tuple[Any, ...] = ()

    class Config:
        """The Pydantic config."""

        # Required to allow tuple fields.
        frozen = True


class EventChain(Base):
    """An event chain specifies a sequence of events."""

    # The list of events.
    events: List[EventSpec]
