"""Pynecone CLI to create, run, and deploy apps."""

import os
import signal
import subprocess
import types

import requests
import typer

from pynecone import constants, utils
from pynecone.compiler import templates

# Create the app.
cli = typer.Typer()


@cli.command()
def init():
    """Initialize a new Pynecone app."""
    # The app name is the same as the name of the current directory.
    app_name = os.getcwd().split(os.path.sep)[-1]
    typer.echo(f"Initializing {app_name}")

    # Install bun if it isn't already installed.
    if utils.which(constants.BUN) is None:
        typer.echo("Installing bun")
        os.system(constants.INSTALL_BUN)

    # Only create the app directory if it doesn't exist.
    if not os.path.exists(constants.CONFIG_FILE):
        # Create a configuration file.
        with open(constants.CONFIG_FILE, "w") as f:
            f.write(templates.PCCONFIG.format(app_name=app_name))

        # Initialize the app directory.
        utils.cp(constants.APP_TEMPLATE_DIR, app_name)
        utils.mv(
            os.path.join(app_name, constants.APP_TEMPLATE_FILE),
            os.path.join(app_name, app_name + constants.PY_EXT),
        )
        utils.mkdir(constants.APP_ASSETS_DIR)

    # Initialize the web directory.
    utils.rm(os.path.join(constants.WEB_TEMPLATE_DIR, constants.NODE_MODULES))
    utils.rm(os.path.join(constants.WEB_TEMPLATE_DIR, constants.PACKAGE_LOCK))
    utils.cp(constants.WEB_TEMPLATE_DIR, constants.WEB_DIR)


@cli.command()
def run():
    """Run the app."""
    app = utils.get_app()
    run_frontend(app)
    run_backend(app)


@cli.command()
def deploy():
    """Deploy the app to the hosting service."""
    # Get the app config.
    pcconfig = utils.get_config()
    pcconfig.API_HOST = utils.get_production_backend_url()

    # Compile the app in production mode.
    typer.echo("Compiling production app")
    app = utils.get_app()
    app.app.compile()

    # Export and zip up the frontend and backend.
    cmd = r"rm -rf .web/_static || true; cd .web && bun run export && cd _static  && zip -r ../../frontend.zip ./* && cd ../.. && zip -r backend.zip ./* -x .web/\* ./assets\* ./frontend.zip\* ./backend.zip\*"
    os.system(cmd)

    data = {"userId": pcconfig.USERNAME, "projectId": pcconfig.APP_NAME}
    original_response = requests.get(
        "https://0v8g446r99.execute-api.us-west-1.amazonaws.com/prod/hosting-api",
        params=data,
    )
    response = original_response.json()
    print("response", response)
    frontend = response["frontend_resources_url"]
    backend = response["backend_resources_url"]

    # Upload the frontend and backend.
    with open("frontend.zip", "rb") as f:
        response = requests.put(frontend, data=f)

    with open("backend.zip", "rb") as f:
        response = requests.put(backend, data=f)


def run_frontend(app: types.ModuleType):
    """Run the frontend for the Pynecone app.

    Args:
        app: The user's app module.
    """
    # Initialize the web directory if it doesn't exist.
    utils.cp(constants.WEB_TEMPLATE_DIR, constants.WEB_DIR, overwrite=False)

    # Install the frontend dependencies.
    subprocess.call(constants.INSTALL_DEPENDENCIES, cwd=constants.WEB_DIR)

    # Link the assets folder.
    utils.ln(
        src=os.path.join("..", constants.APP_ASSETS_DIR), dest=constants.WEB_ASSETS_DIR
    )

    # Compile the frontend.
    app.app.compile()

    # Run the development server in the background.
    frontend_process = subprocess.Popen(constants.RUN_FRONTEND, cwd=constants.WEB_DIR)

    def stop_frontend(*_):
        """Stop the frontend.

        Args:
            _: Ignored arguments passed to the signal handler.
        """
        utils.kill(frontend_process.pid)

    # Add signal handlers to stop the frontend.
    signal.signal(signal.SIGINT, stop_frontend)
    signal.signal(signal.SIGTERM, stop_frontend)


def run_backend(app: types.ModuleType):
    """Run the backend for the Pynecone app.

    Args:
        app: The user's app module.
    """
    subprocess.call(constants.RUN_BACKEND + [f"{app.__name__}:{constants.API_VAR}"])


def main():
    """Run the CLI app."""
    cli()


if __name__ == "__main__":
    main()
