"""Define a state property."""
from __future__ import annotations

import json
from abc import ABC
from typing import _GenericAlias  # type: ignore
from typing import TYPE_CHECKING, Any, Callable, Optional, Type, Union

from pynecone import constants, utils
from pynecone.base import Base

if TYPE_CHECKING:
    from pynecone.state import State


class Property(ABC):
    """An abstract property."""

    # The name of the property.
    name: str

    # The type of the property.
    type_: Type

    # The name of the enclosing state.
    state: str = ""

    # Whether this is a local javascript variable.
    is_local: bool = False

    @classmethod
    def create(cls, value: Any) -> Optional[Property]:
        """Create a property from a value.

        Args:
            value: The value to create the property from.

        Returns:
            The property.
        """
        # Check for none values.
        if value is None:
            return None

        # If the value is already a property, do nothing.
        if isinstance(value, Property):
            return value

        name = json.dumps(value) if not isinstance(value, str) else value

        return BaseProperty(name=name, type_=type(value), is_local=True)

    @classmethod
    def __class_getitem__(cls, type_: str) -> _GenericAlias:
        """Get a typed property.

        Args:
            type_: The type of the property.

        Returns:
            The property class item.
        """
        return _GenericAlias(cls, type_)

    def equals(self, other: Property) -> bool:
        """Check if two properties are equal.

        Args:
            other: The other property to compare.

        Returns:
            Whether the properties are equal.
        """
        return (
            self.name == other.name
            and self.type_ == other.type_
            and self.state == other.state
            and self.is_local == other.is_local
        )

    def __hash__(self) -> int:
        """Define a hash function for a property.

        Returns:
            The hash of the property.
        """
        return hash((self.name, str(self.type_)))

    def __str__(self) -> str:
        """Wrap the property so it can be used in templates.

        Returns:
            The wrapped property, i.e. {state.property}.
        """
        if self.is_local:
            return self.full_name
        return utils.wrap(self.full_name, "{")

    def __getitem__(self, i) -> Property:
        """Index into a property.

        Args:
            i: The index to index into.

        Returns:
            The indexed property.
        """
        # The type of the indexed property.
        type_ = str

        # Convert any properties to local properties.
        if isinstance(i, Property):
            i = BaseProperty(
                name=i.name, type_=i.type_, state=self.state, is_local=True
            )

        if utils._issubclass(self.type_, Union[tuple, list]):
            assert isinstance(
                i, utils.get_args(Union[int, Property])
            ), "Index must be an integer."
            if isinstance(self.type_, _GenericAlias):
                type_ = utils.get_args(self.type_)[0]
            else:
                type_ = Any
        elif utils._issubclass(self.type_, dict):
            if isinstance(i, str):
                i = utils.wrap(i, '"')
            if isinstance(self.type_, _GenericAlias):
                type_ = utils.get_args(self.type_)[1]
            else:
                type_ = Any
        else:
            raise TypeError("Property does not support indexing.")

        return BaseProperty(
            name=f"{self.name}[{i}]",
            type_=type_,
            state=self.state,
        )

    def __getattribute__(self, name: str) -> Property:
        """Get a property attribute.

        Args:
            name: The name of the attribute.

        Returns:
            The property attribute.
        """
        try:
            return super().__getattribute__(name)
        except Exception as e:
            # Check if the attribute is one of the class fields.
            if (
                not name.startswith("_")
                and hasattr(self.type_, "__fields__")
                and name in self.type_.__fields__
            ):
                return BaseProperty(
                    name=f"{self.name}.{name}",
                    type_=self.type_.__fields__[name],
                    state=self.state,
                )
            raise e

    def operation(
        self,
        op: str = "",
        other: Optional[Property] = None,
        type_: Optional[Type] = None,
        flip: bool = False,
        fn: Optional[str] = None,
    ) -> Property:
        """Perform an operation on a property.

        Args:
            op: The operation to perform.
            other: The other property to perform the operation on.
            type_: The type of the operation result.
            flip: Whether to flip the order of the operation.
            fn: A function to apply to the operation.

        Returns:
            The operation result.
        """
        other = Property.create(other)
        if type_ is None:
            type_ = self.type_
        if other is None:
            name = f"{op}{self.full_name}"
        else:
            props = (self, other) if not flip else (other, self)
            name = f"{props[0].full_name} {op} {props[1].full_name}"
            if fn is None:
                name = utils.wrap(name, "(")
        if fn is not None:
            name = f"{fn}({name})"
        return BaseProperty(
            name=name,
            type_=type_,
        )

    def compare(self, op: str, other: Property) -> Property:
        """Compare two properties with inequalities.

        Args:
            op: The comparison operator.
            other: The other property to compare with.

        Returns:
            The comparison result.
        """
        return self.operation(op, other, bool)

    def __invert__(self) -> Property:
        """Invert a property.

        Returns:
            The inverted property.
        """
        return self.operation("!", type_=bool)

    def __neg__(self) -> Property:
        """Negate a property.

        Returns:
            The negated property.
        """
        return self.operation(fn="-")

    def __abs__(self) -> Property:
        """Get the absolute value of a property.

        Returns:
            A property with the absolute value.
        """
        return self.operation(fn="Math.abs")

    def __eq__(self, other: Property) -> Property:
        """Perform an equality comparison.

        Args:
            other: The other property to compare with.

        Returns:
            A property representing the equality comparison.
        """
        return self.compare("==", other)

    def __ne__(self, other: Property) -> Property:
        """Perform an inequality comparison.

        Args:
            other: The other property to compare with.

        Returns:
            A property representing the inequality comparison.
        """
        return self.compare("!=", other)

    def __gt__(self, other: Property) -> Property:
        """Perform a greater than comparison.

        Args:
            other: The other property to compare with.

        Returns:
            A property representing the greater than comparison.
        """
        return self.compare(">", other)

    def __ge__(self, other: Property) -> Property:
        """Perform a greater than or equal to comparison.

        Args:
            other: The other property to compare with.

        Returns:
            A property representing the greater than or equal to comparison.
        """
        return self.compare(">=", other)

    def __lt__(self, other: Property) -> Property:
        """Perform a less than comparison.

        Args:
            other: The other property to compare with.

        Returns:
            A property representing the less than comparison.
        """
        return self.compare("<", other)

    def __le__(self, other: Property) -> Property:
        """Perform a less than or equal to comparison.

        Args:
            other: The other property to compare with.

        Returns:
            A property representing the less than or equal to comparison.
        """
        return self.compare("<=", other)

    def __add__(self, other: Property) -> Property:
        """Add two properties.

        Args:
            other: The other property to add.

        Returns:
            A property representing the sum.
        """
        return self.operation("+", other)

    def __radd__(self, other: Property) -> Property:
        """Add two properties.

        Args:
            other: The other property to add.

        Returns:
            A property representing the sum.
        """
        return self.operation("+", other, flip=True)

    def __sub__(self, other: Property) -> Property:
        """Subtract two properties.

        Args:
            other: The other property to subtract.

        Returns:
            A property representing the difference.
        """
        return self.operation("-", other)

    def __rsub__(self, other: Property) -> Property:
        """Subtract two properties.

        Args:
            other: The other property to subtract.

        Returns:
            A property representing the difference.
        """
        return self.operation("-", other, flip=True)

    def __mul__(self, other: Property) -> Property:
        """Multiply two properties.

        Args:
            other: The other property to multiply.

        Returns:
            A property representing the product.
        """
        return self.operation("*", other)

    def __rmul__(self, other: Property) -> Property:
        """Multiply two properties.

        Args:
            other: The other property to multiply.

        Returns:
            A property representing the product.
        """
        return self.operation("*", other, flip=True)

    def __pow__(self, other: Property) -> Property:
        """Raise a property to a power.

        Args:
            other: The power to raise to.

        Returns:
            A property representing the power.
        """
        return self.operation(",", other, fn="Math.pow")

    def __rpow__(self, other: Property) -> Property:
        """Raise a property to a power.

        Args:
            other: The power to raise to.

        Returns:
            A property representing the power.
        """
        return self.operation(",", other, flip=True, fn="Math.pow")

    def __truediv__(self, other: Property) -> Property:
        """Divide two properties.

        Args:
            other: The other property to divide.

        Returns:
            A property representing the quotient.
        """
        return self.operation("/", other)

    def __rtruediv__(self, other: Property) -> Property:
        """Divide two properties.

        Args:
            other: The other property to divide.

        Returns:
            A property representing the quotient.
        """
        return self.operation("/", other, flip=True)

    def __floordiv__(self, other: Property) -> Property:
        """Divide two properties.

        Args:
            other: The other property to divide.

        Returns:
            A property representing the quotient.
        """
        return self.operation("/", other, fn="Math.floor")

    def __mod__(self, other: Property) -> Property:
        """Get the remainder of two properties.

        Args:
            other: The other property to divide.

        Returns:
            A property representing the remainder.
        """
        return self.operation("%", other)

    def __rmod__(self, other: Property) -> Property:
        """Get the remainder of two properties.

        Args:
            other: The other property to divide.

        Returns:
            A property representing the remainder.
        """
        return self.operation("%", other, flip=True)

    def __and__(self, other: Property) -> Property:
        """Perform a logical and.

        Args:
            other: The other property to perform the logical and with.

        Returns:
            A property representing the logical and.
        """
        return self.operation("&&", other)

    def __rand__(self, other: Property) -> Property:
        """Perform a logical and.

        Args:
            other: The other property to perform the logical and with.

        Returns:
            A property representing the logical and.
        """
        return self.operation("&&", other, flip=True)

    def __or__(self, other: Property) -> Property:
        """Perform a logical or.

        Args:
            other: The other property to perform the logical or with.

        Returns:
            A property representing the logical or.
        """
        return self.operation("||", other)

    def __ror__(self, other: Property) -> Property:
        """Perform a logical or.

        Args:
            other: The other property to perform the logical or with.

        Returns:
            A property representing the logical or.
        """
        return self.operation("||", other, flip=True)

    def foreach(self, fn: Callable) -> Property:
        """Return a list of components. after doing a foreach on this property.

        Args:
            fn: The function to call on each component.

        Returns:
            A property representing foreach operation.
        """
        arg = BaseProperty(
            name=utils.get_unique_variable_name(),
            type_=self.type_,
        )
        return BaseProperty(
            name=f"{self.full_name}.map(({arg.name}, i) => {fn(arg, key='i')})",
            type_=self.type_,
        )

    def to(self, type_: Type) -> Property:
        """Convert the type of the property.

        Args:
            type_: The type to convert to.

        Returns:
            The converted property.
        """
        return BaseProperty(
            name=self.name,
            type_=type_,
            state=self.state,
            is_local=self.is_local,
        )

    @property
    def full_name(self) -> str:
        """Get the full name of the property.

        Returns:
            The full name of the property.
        """
        if self.state == "":
            return self.name
        return ".".join([self.state, self.name])

    def set_state(self, state: Type[State]) -> Any:
        """Set the state of the property.

        Args:
            state: The state to set.

        Returns:
            The property with the set state.
        """
        self.state = state.get_full_name()
        return self


class BaseProperty(Property, Base):
    """A base (non-computed) property of the app state."""

    # The name of the property.
    name: str

    # The type of the property.
    type_: Any

    # The name of the enclosing state.
    state: str = ""

    # Whether this is a local javascript variable.
    is_local: bool = False

    def __hash__(self) -> int:
        """Define a hash function for a property.

        Returns:
            The hash of the property.
        """
        return hash((self.name, str(self.type_)))

    def get_default_value(self) -> Any:
        """Get the default value of the property.

        Returns:
            The default value of the property.
        """
        if isinstance(self.type_, _GenericAlias):
            type_ = self.type_.__origin__
        else:
            type_ = self.type_
        if issubclass(type_, str):
            return ""
        if issubclass(type_, utils.get_args(Union[int, float])):
            return 0
        if issubclass(type_, bool):
            return False
        if issubclass(type_, list):
            return []
        if issubclass(type_, dict):
            return {}
        if issubclass(type_, tuple):
            return ()
        if issubclass(type_, set):
            return set()
        return None

    def get_setter_name(self) -> str:
        """Get the name of the property's generated setter function.

        Returns:
            The name of the setter function.
        """
        setter = constants.SETTER_PREFIX + self.name
        if self.state == "":
            return setter
        return ".".join((self.state, setter))

    def get_setter(self) -> Callable[[State], None]:
        """Get the property's setter function.

        Returns:
            A function that that creates a setter for the property.
        """

        def setter(state: State, **kwargs):
            """Get the setter for the property.

            Args:
                state: The state within which we add the setter function.
                **kwargs: The values to set - the kwargs should have length 1.
            """
            setattr(state, self.name, next(iter(kwargs.values())))

        setter.__qualname__ = self.get_setter_name()

        return setter

    def json(self) -> str:
        """Convert the object to a json string.

        Returns:
            The object as a json string.
        """
        return self.__config__.json_dumps(self.dict())


class ComputedProperty(property, Property):
    """A field with computed getters."""

    @property
    def name(self) -> str:
        """Get the name of the property.

        Returns:
            The name of the property.
        """
        assert self.fget is not None, "Property must have a getter."
        return self.fget.__name__

    @property
    def type_(self):
        """Get the type of the property.

        Returns:
            The type of the property.
        """
        return self.fget.__annotations__["return"]
