"""Define the pynecone state specification."""
from __future__ import annotations

import functools
from abc import ABC
from typing import Any, Callable, ClassVar, Dict, Optional, Sequence, Set, Type

from pynecone import utils
from pynecone.base import Base
from pynecone.event import Event, EventHandler
from pynecone.property import BaseProperty, ComputedProperty, Property

Delta = Dict[str, Any]


class State(Base, ABC):
    """The state of the app."""

    # A map from the property name to the property.
    properties: ClassVar[Dict[str, Property]] = {}

    # The base properties of the class.
    base_properties: ClassVar[Dict[str, BaseProperty]] = {}

    # The computed properties of the class.
    computed_properties: ClassVar[Dict[str, ComputedProperty]] = {}

    # Properties inherited by the parent state.
    inherited_properties: ClassVar[Dict[str, Property]] = {}

    # The parent state.
    parent_state: Optional[State] = None

    # The substates of the state.
    substates: Dict[str, State] = {}

    # The set of dirty properties.
    dirty_properties: Set[str] = set()

    # The set of dirty substates.
    dirty_substates: Set[str] = set()

    # The current page.
    page: str = ""

    def __init__(self, *args, **kwargs):
        """Initialize the state.

        Args:
            *args: The args to pass to the pydantic init method.
            **kwargs: The kwargs to pass to the pydantic init method.
        """
        super().__init__(*args, **kwargs)

        # Setup the substates.
        for substate in self.get_substates():
            self.substates[substate.get_name()] = substate().set(parent_state=self)

    def __repr__(self) -> str:
        """Get the string representation of the state.

        Returns:
            The string representation of the state.
        """
        return f"{self.__class__.__name__}({self.dict()})"

    @classmethod
    def __init_subclass__(cls, **kwargs):
        """Do some magic for the subclass initialization.

        Args:
            **kwargs: The kwargs to pass to the pydantic init_subclass method.
        """
        super().__init_subclass__(**kwargs)

        # Get the parent properties.
        parent_state = cls.get_parent_state()
        if parent_state is not None:
            cls.inherited_properties = parent_state.properties

        # Set the base and computed properties.
        skip_properties = set(cls.inherited_properties) | {
            "parent_state",
            "substates",
            "dirty_properties",
            "dirty_substates",
        }
        cls.base_properties = {
            f.name: BaseProperty(name=f.name, type_=f.outer_type_).set_state(cls)
            for f in cls.get_fields().values()
            if f.name not in skip_properties
        }
        cls.computed_properties = {
            v.name: v.set_state(cls)
            for v in cls.__dict__.values()
            if isinstance(v, ComputedProperty)
        }
        cls.properties = {
            **cls.inherited_properties,
            **cls.base_properties,
            **cls.computed_properties,
        }

        # Setup the base properties at the class level.
        for prop in cls.base_properties.values():
            cls._set_property(prop)
            cls._create_setter(prop)
            cls._set_default_value(prop)

        # Set up the event handlers.
        events = {
            name: fn
            for name, fn in cls.__dict__.items()
            if not name.startswith("_") and isinstance(fn, Callable)
        }
        if parent_state is None and "hydrate" not in events:
            events["hydrate"] = lambda self: None
        for name, fn in events.items():
            event_handler = EventHandler(fn=fn)
            setattr(cls, name, event_handler)

    @classmethod
    @functools.lru_cache()
    def get_parent_state(cls) -> Optional[Type[State]]:
        """Get the parent state.

        Returns:
            The parent state.
        """
        parent_states = [
            base
            for base in cls.__bases__
            if issubclass(base, State) and base is not State
        ]
        assert len(parent_states) < 2, "Only one parent state is allowed."
        return parent_states[0] if len(parent_states) == 1 else None

    @classmethod
    @functools.lru_cache()
    def get_substates(cls) -> Set[Type[State]]:
        """Get the substates of the state.

        Returns:
            The substates of the state.
        """
        return {subclass for subclass in cls.__subclasses__()}

    @classmethod
    @functools.lru_cache()
    def get_name(cls) -> str:
        """Get the name of the state.

        Returns:
            The name of the state.
        """
        return utils.to_snake_case(cls.__name__)

    @classmethod
    @functools.lru_cache()
    def get_full_name(cls) -> str:
        """Get the full name of the state.

        Returns:
            The full name of the state.
        """
        name = cls.get_name()
        parent_state = cls.get_parent_state()
        if parent_state is not None:
            name = ".".join((parent_state.get_full_name(), name))
        return name

    @classmethod
    @functools.lru_cache()
    def get_class_substate(cls, path: Sequence[str]) -> Type[State]:
        """Get the class substate.

        Args:
            path: The path to the substate.

        Returns:
            The class substate.
        """
        if len(path) == 0:
            return cls
        if path[0] == cls.get_name():
            if len(path) == 1:
                return cls
            path = path[1:]
        for substate in cls.get_substates():
            if path[0] == substate.get_name():
                return substate.get_class_substate(path[1:])
        raise ValueError(f"Invalid path: {path}")

    @classmethod
    def get_class_property(cls, path: Sequence[str]) -> Any:
        """Get the class property.

        Args:
            path: The path to the property.

        Returns:
            The class property.
        """
        path, name = path[:-1], path[-1]
        substate = cls.get_class_substate(tuple(path))
        if not hasattr(substate, name):
            raise ValueError(f"Invalid path: {path}")
        return getattr(substate, name)

    @classmethod
    def _set_property(cls, prop: BaseProperty):
        """Set the property as a class member.

        Args:
            prop: The property instance to set.
        """
        setattr(cls, prop.name, prop)

    @classmethod
    def _create_setter(cls, prop: BaseProperty):
        """Create a setter for the property.

        Args:
            prop: The property to create a setter for.
        """
        setter_name = prop.get_setter_name().split(".")[-1]
        if setter_name not in cls.__dict__:
            setattr(cls, setter_name, prop.get_setter())

    @classmethod
    def _set_default_value(cls, prop: BaseProperty):
        """Set the default value for the property.

        Args:
            prop: The property to set the default value for.
        """
        # Get the pydantic field for the property.
        field = cls.get_fields()[prop.name]
        default_value = prop.get_default_value()
        if field.required and default_value is not None:
            field.required = False
            field.default = default_value

    def __getattribute__(self, name: str) -> Any:
        """Get the attribute.

        Args:
            name: The name of the attribute.

        Returns:
            The attribute.
        """
        # If it is an inherited property, return from the parent state.
        if name != "inherited_properties" and name in self.inherited_properties:
            return getattr(self.parent_state, name)
        try:
            return super().__getattribute__(name)
        except Exception as e:
            # Check if the attribute is a substate.
            if name in self.substates:
                return self.substates[name]
            raise e

    def __setattr__(self, name: str, value: Any):
        """Set the attribute.

        Args:
            name: The name of the attribute.
            value: The value of the attribute.
        """
        if name != "inherited_properties" and name in self.inherited_properties:
            setattr(self.parent_state, name, value)
            return

        # Set the attribute.
        super().__setattr__(name, value)

        # Add the property to the dirty list.
        if name in self.properties:
            self.dirty_properties.add(name)
            self.mark_dirty()

    def reset(self):
        """Reset all the base properties to their default values."""
        # Reset the base properties.
        fields = self.get_fields()
        for prop_name in self.base_properties:
            setattr(self, prop_name, fields[prop_name].default)

        # Recursively reset the substates.
        for substate in self.substates.values():
            substate.reset()

        # Clean the state.
        self.clean()

    def get_substate(self, path: Sequence[str]) -> Optional[State]:
        """Get the substate.

        Args:
            path: The path to the substate.

        Returns:
            The substate.
        """
        if len(path) == 0:
            return self
        if path[0] == self.get_name():
            if len(path) == 1:
                return self
            path = path[1:]
        if path[0] not in self.substates:
            raise ValueError(f"Invalid path: {path}")
        return self.substates[path[0]].get_substate(path[1:])

    def process(self, event: Event) -> Delta:
        """Process an event.

        Args:
            event: The event to process.

        Returns:
            The state delta after processing the event.
        """
        # Get the event handler.
        path = event.name.split(".")
        path, name = path[:-1], path[-1]
        substate = self.get_substate(path)
        handler = getattr(substate, name)

        # Process the event.
        fn = functools.partial(handler.fn, substate)
        fn(**event.payload)

        # Return the substate and the delta.
        return self.get_delta()

    def get_delta(self) -> Delta:
        """Get the delta for the state.

        Returns:
            The delta for the state.
        """
        if len(self.dirty_properties) == 0:
            delta = {}
        else:
            delta = {
                self.get_full_name(): {
                    prop: getattr(self, prop)
                    for prop in self.dirty_properties
                    | set(self.computed_properties.keys())
                }
            }
        for substate in self.dirty_substates:
            delta.update(self.substates[substate].get_delta())
        return delta

    def mark_dirty(self):
        """Mark the substate and all parent states as dirty."""
        if self.parent_state is not None:
            self.parent_state.dirty_substates.add(self.get_name())
            self.parent_state.mark_dirty()

    def clean(self):
        """Reset the dirty properties."""
        for substate in self.dirty_substates:
            self.substates[substate].clean()
        self.dirty_properties = set()
        self.dirty_substates = set()

    def dict(self, include_computed: bool = True, **kwargs) -> Dict[str, Any]:
        """Convert the object to a dictionary.

        Args:
            include_computed: Whether to include computed properties.
            **kwargs: Kwargs to pass to the pydantic dict method.

        Returns:
            The object as a dictionary.
        """
        base_properties = {
            prop_name: self.get_value(getattr(self, prop_name))
            for prop_name in self.base_properties
        }
        computed_properties = (
            {
                # Include the computed properties.
                prop_name: self.get_value(getattr(self, prop_name))
                for prop_name in self.computed_properties
            }
            if include_computed
            else {}
        )
        substate_properties = {
            k: v.dict(include_computed=include_computed, **kwargs)
            for k, v in self.substates.items()
        }
        properties = {**base_properties, **computed_properties, **substate_properties}
        return {k: properties[k] for k in sorted(properties)}
